#!/usr/bin/perl
#
# Copyright (c) 2005 IDEALX
#
# This software is under GNU PUBLIC LICENSE VERSION 2.
#
use Getopt::Long;
use strict;

sub usage {
    print "gen_rpm_base.pl , Copyright (C) 2003 IDEALX (http://IDEALX.org/)\n\n";
    print "\n";
    print ("Usage: gen_rpm_base.pl\n\t",
           "Available options:\n\t",
           "--comps <path>\n\t",
           "--destpath <path>\n\t",
           "--source <path> (RPM sources, at least one needed)\n\t",
           "[--force <path>] (mandatory sources)\n\t",
           "[--debug] (dont copy anything, just print what should be done)\n\t",
           "[--suse] (for SuSE, RedHat by default)\n\t",
           "[--verbose] (verbose output)\n");
    "[-h, --help] (this help text)\n\t",
        exit(1);
}

my @sources = ();
my @force = ();
my $target = "i686";
my $comps;
my $dest;
my $help = 0;
my $debug = 0;
my $verbose = 0;
my $suse = 0;

GetOptions ("destpath=s" => \$dest, "comps=s" => \$comps,
            "target=s" => \$target, "h|help" => \$help, "debug" => \$debug,
            "verbose" => \$verbose,
            "source=s" => \@sources, "force=s" => \@force, "suse" => \$suse) || die "RTFS";

&usage if $help || $Getopt::Long::error;
&usage unless($dest and $comps and scalar(@sources));

sub getBest {
    my ($v1, $v2, @order) = @_;
    foreach my $w (@order) {
        if ($v1 eq $w) {
            return 1;
        } elsif ($v2 eq $w) {
            return -1;
        }
    }
    return 0;
}

sub CompareVersions {
    my ($v1, $v2)= @_;
    return 0 if($v1 eq $v2);

    my @sv1 = split m/[.]/, $v1;
    my @sv2 = split m/[.]/, $v2;

    while (scalar(@sv1) and scalar(@sv2)) {
        my $a = shift(@sv1);
        my $b = shift(@sv2);

        if ($a < $b) {
            return -1;
        } elsif ($a > $b) {
            return 1;
        }
    }

    if (scalar(@sv1)) {
        return 1;
    } else {
        return -1;
    }
}

sub bestPackage {
    my ($version, $release, $arch, $dir, $file, $oversion,
        $orelease, $oarch, $odir, $ofile, @order) = @_;
    my $choice = 1;

    my $cv = CompareVersions($version, $oversion);

    if ($cv > 0) {
        $choice = 0;
    } elsif ($cv == 0) {
        my $cr = CompareVersions($release, $orelease);
        if ($cr > 0) {
            $choice = 0;
        } elsif ($cr == 0) {
            $a = getBest( $arch, $oarch, @order);
            if ($a == 1) {
                $choice = 0;
            }
        }
    }

    if ($choice == 1) {
        return ($oversion, $orelease, $oarch, $odir, $ofile);
    } else {
        return ($version, $release, $arch, $dir, $file);
    }
}

sub read_repository {
    my ($dir, $force, $arch) = @_;
    my %ret;
    my @rpms;
    if (-f "$dir/cache.txt") {
        print "Using cache: $dir/cache.txt\n" if $verbose;
        open (CACHE, "$dir/cache.txt");
        @rpms = <CACHE>;
        close(CACHE);
    } else {
        @rpms = split "\n",`rpm --queryformat "%{NAME} %{VERSION} %{RELEASE} %{ARCH} %{NAME}-%{VERSION}-%{RELEASE}.%{ARCH}.rpm\n" -qp $dir/*.rpm`;
    }

    my %packages;
    my %versions;
    my %release;
    my %archs;
    my %files;
    while (@rpms) {
        my ($name, $version, $release, $arch, $file) = split ' ', shift @rpms;
        my $dir;
        # Remove all files except for the best arch ($target > i386 > noarch)
        if ($packages{$name}) {
            ($version, $release, $arch, $dir, $file) = bestPackage
                ($version, $release, $arch, "", $file,
                 $versions{$name}, $release{$name}, $archs{$name}, "", $files{$name},
                 ($target, "i386", "noarch"));
        }
        $packages{$name} = 1;
        $versions{$name} = $version;
        $release{$name} = $release;
        $archs{$name} = $arch;
        $files{$name} = $file;
    }

    $ret{directory} = $dir;
    $ret{force} = $force;
    $ret{packages} = \%packages;
    $ret{versions} = \%versions;
    $ret{release} = \%release;
    $ret{archs} = \%archs;
    $ret{files} = \%files;

    return \%ret;
}

sub load_comps {
    my ($file) = @_;
    my %rpm;

    open (COMPS,"< $file") or die "Can't open $file";

    while (<COMPS>) {           # parse comps file
        print "parsing $_" if $verbose;
        if (/^\s*(\d+) (\w+)/) { # start block
            print  "#remove start block ...\n" if $debug;
        } elsif (/^\s*$/) {		# blank line
            print  "#remove blank ...\n" if $debug;
        } elsif (/\}/) {		# end block
            print  "#remove block ...\n" if $debug;
        } elsif (/^\d+\s*$/) { 	# version number
            print  "#remove number ...\n" if $debug;
        } elsif (/^\d+\s+/) { 	# version number
            print  "#remove number 2 ...\n" if $debug;
        } elsif (/^\s+@/) {		# cat name
            print  "#remove name ...\n" if $debug;
        } elsif (/\?/) { 		# cat name
            print  "#remove name 2 ...\n" if $debug;
        } elsif (/\([^\)]+\):\s+(\S+)/) {
            print  "#Adding rpm $1 ...\n" if $verbose;
            $rpm{$1}=1;
        } elsif (/[^:]+\:\s+(\S+)/) {
            print  "#Adding rpm $1 ...\n" if $verbose;
            $rpm{$1}=1;
        } elsif (/\s+(\S+)/) {
            print  "#Adding rpm $1 ...\n" if $verbose;
            $rpm{$1}=1;
        } else {
            print STDERR "arg, unknown ! $_ \n";
        }
    }

    close COMPS;
    return keys(%rpm);
}

print "Stage 1: Reading comps file $comps\n";

my @rpms = load_comps($comps);

my @repositories;

print "Stage 2: Reading Repositories\n";

foreach my $source (@sources) {
    print "Loading $source\n";
    push @repositories, read_repository($source, $target, 0);
}

foreach my $source (@force) {
    print "Loading $source\n";
    push @repositories, read_repository($source, $target, 1);
}

print "Stage 3: Computing Best packages\n";

my %packages;
my %versions;
my %release;
my %archs;
my %dirs;
my %files;
my $missing = 0;
foreach my $rpm (@rpms) {
    print "Package $rpm is needed\n" if $verbose;

    foreach my $repository (@repositories) {
        if ($repository->{packages}->{$rpm}) {
            if ($packages{$rpm}) {
                # Choose best suited package
                my ($version, $release, $arch, $dir, $file);
	
                if ($repository->{force}) {
                    # This repository is marked as forced...
                    ($version, $release, $arch, $dir, $file) = ($repository->{versions}->{$rpm},
                                                                $repository->{release}->{$rpm},
                                                                $repository->{archs}->{$rpm},
                                                                $repository->{directory},
                                                                $repository->{files}->{$rpm});
                } else {
                    ($version, $release, $arch, $dir, $file) = bestPackage
                        ($repository->{versions}->{$rpm}, $repository->{release}->{$rpm},
                         $repository->{archs}->{$rpm}, $repository->{directory},
                         $repository->{files}->{$rpm},
                         $versions{$rpm}, $release{$rpm}, $archs{$rpm}, $dirs{$rpm}, $files{$rpm},
                         ($target, "i386", "noarch"));
                }
	
                $versions{$rpm} = $version;
                $release{$rpm} = $release;
                $archs{$rpm} = $arch;
                $dirs{$rpm} = $dir;
                $files{$rpm} = $file;
            } else {
                $packages{$rpm} = 1;
                $versions{$rpm} = $repository->{versions}->{$rpm};
                $release{$rpm} = $repository->{release}->{$rpm};
                $archs{$rpm} = $repository->{archs}->{$rpm};
                $dirs{$rpm} = $repository->{directory};
                $files{$rpm} = $repository->{files}->{$rpm};
            }
        }
    }
    unless ($packages{$rpm}) {
        print "Error package $rpm is missing\n";
        $missing = 1;
    }
}

die "Generation failed due to missing package\n" if $missing;

print "Stage 4: Generating Final repository\n";

if ($suse) {
	if (! -d $dest."/suse") {
		print "Creating directory $dest/suse)\n" if $verbose;
		mkdir($dest."/suse");
	}

	for my $dir ("i586", "i686", "noarch") {
		my $path2dir = "$dest/suse/$dir";
		if (! -d $path2dir) {
			print "Creating directory $path2dir\n" if $verbose;
			mkdir($path2dir) or die "Unable to create directory $path2dir: $!\n";
		}
	}
}

foreach my $rpm (@rpms) {
    my ($file, $source) = ($files{$rpm}, $dirs{$rpm});
    print "$source/$file\t=>\t$dest/$file\n" if $verbose;
    if ($suse) {  
        link("$source/$file", "$dest/suse/$archs{$rpm}/$file");
    } else { 
        link("$source/$file", "$dest/$file"); 
    }
}

exit(0);

